##############################################################################
#
# Copyright (C) Zenoss, Inc. 2013-2017, all rights reserved.
#
# This content is made available according to terms specified in
# License.zenoss under the directory where your Zenoss product is installed.
#
##############################################################################

import logging
log = logging.getLogger('zen.AWS')

import calendar
import datetime

from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks

from zope.component import adapts
from zope.interface import implements

from Products.ZenEvents import ZenEventClasses
from Products.Zuul.form import schema
from Products.Zuul.infos import ProxyProperty
from Products.Zuul.infos.template import RRDDataSourceInfo
from Products.Zuul.interfaces import IRRDDataSourceInfo
from Products.Zuul.utils import ZuulMessageFactory as _t

from ZenPacks.zenoss.PythonCollector.datasources.PythonDataSource \
    import PythonDataSource

from ZenPacks.zenoss.AWS.utils import addLocalLibPath
addLocalLibPath()

import boto
import boto.ec2
import boto.ec2.cloudwatch

from ZenPacks.zenoss.AWS.txboto import TwistedCloudWatchConnection
from ZenPacks.zenoss.AWS.dsplugins import AWSBasePlugin


class AmazonCloudWatchDataSource(PythonDataSource):
    '''
    Datasource used to capture datapoints from Amazon CloudWatch.
    '''

    ZENPACKID = 'ZenPacks.zenoss.AWS'

    sourcetypes = ('Amazon CloudWatch',)
    sourcetype = sourcetypes[0]

    # RRDDataSource
    component = '${here/id}'
    cycletime = 300

    # PythonDataSource
    plugin_classname = 'ZenPacks.zenoss.AWS.datasources.'\
        'AmazonCloudWatchDataSource.AmazonCloudWatchDataSourcePlugin'

    # AmazonCloudWatchDataSource
    namespace = ''
    metric = ''
    statistic = 'Average'
    dimension = '${here/getDimension}'
    region = '${here/getRegionId}'

    _properties = PythonDataSource._properties + (
        {'id': 'namespace', 'type': 'string'},
        {'id': 'metric', 'type': 'string'},
        {'id': 'statistic', 'type': 'string'},
        {'id': 'dimension', 'type': 'string'},
        {'id': 'region', 'type': 'string'},
        )


class IAmazonCloudWatchDataSourceInfo(IRRDDataSourceInfo):
    '''
    API Info interface for AmazonCloudWatchDataSource.
    '''

    namespace = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Namespace'))

    metric = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Metric Name'))

    statistic = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Statistic'))

    dimension = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Dimension'))

    region = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Region'))


class AmazonCloudWatchDataSourceInfo(RRDDataSourceInfo):
    '''
    API Info adapter factory for AmazonCloudWatchDataSource.
    '''

    implements(IAmazonCloudWatchDataSourceInfo)
    adapts(AmazonCloudWatchDataSource)

    testable = False

    cycletime = ProxyProperty('cycletime')

    namespace = ProxyProperty('namespace')
    metric = ProxyProperty('metric')
    statistic = ProxyProperty('statistic')
    dimension = ProxyProperty('dimension')
    region = ProxyProperty('region')


class AmazonCloudWatchDataSourcePlugin(AWSBasePlugin):
    '''
    Python plugin used to capture datapoints from Amazon CloudWatch.
    '''

    proxy_attributes = (
        'ec2accesskey', 'ec2secretkey', 'zAWSCloudWatchSSL'
        )

    @classmethod
    def params(cls, datasource, context):
        return {
            'namespace': datasource.talesEval(datasource.namespace, context),
            'metric': datasource.talesEval(datasource.metric, context),
            'statistic': datasource.talesEval(datasource.statistic, context),
            'dimension': datasource.talesEval(datasource.dimension, context),
            'region': datasource.talesEval(datasource.region, context),
            }

    def ds_description(self, config, ds):
        return '[%s/%s: %s %s/%s for %s]' % (
            config.id,
            ds.params['region'],
            ds.params['statistic'],
            ds.params['namespace'],
            ds.params['metric'],
            ds.params['dimension'] or 'region')

    @inlineCallbacks
    def collect(self, config):
        data = self.new_data()
        deferreds = []
        results = []
        names = []

        # Usage on device level, may be used for billing/estimated charges.
        if not config.datasources[0].component:
            log.info("Amazon CloudWatch data source used on device level")

        for ds in config.datasources:
            # CloudWatch only accepts periods that are evenly divisible by 60.
            cycletime = (ds.cycletime / 60) * 60
            end_time = datetime.datetime.utcnow()
            start_time = end_time - datetime.timedelta(seconds=4*cycletime)

            dimensions = {}
            if ds.params['dimension']:
                dim_group = ds.params['dimension'].split(';')
                for dim_name, dim_value in [x.split('=') for x in dim_group]:
                    dimensions[dim_name] = dim_value

            con = TwistedCloudWatchConnection.adapt(
                boto.ec2.cloudwatch.connect_to_region(
                    ds.params['region'],
                    aws_access_key_id=ds.ec2accesskey,
                    aws_secret_access_key=ds.ec2secretkey,
                    is_secure=ds.zAWSCloudWatchSSL
            ))

            log.debug("GetMetricStatistics [%s] StartTime=%s, EndTime=%s",
                self.ds_description(config, ds), start_time, end_time)

            d = con.get_metric_statistics(
                cycletime,
                start_time,
                end_time,
                ds.params['metric'],
                ds.params['namespace'],
                ds.params['statistic'],
                dimensions=dimensions
            )
            deferreds.append(d)
            names.append(ds)

        # allow all the requests to complete.
        results = yield defer.DeferredList(deferreds, consumeErrors=True)

        for i, value in enumerate(results):
            success, result = value
            ds = names[i]

            if not success:
                log.error("GetMetricStatistics error [%s]: %s",
                    self.ds_description(config, ds), result.getErrorMessage())
                # Give up, since some of the data couldn't be retrieved.
                continue

            if result:
                log.debug("GetMetricStatistics result [%s]: %s",
                    self.ds_description(config, ds), result)

                data['values'][ds.component].setdefault(ds.datasource, [])
                for item in result:
                    timestamp, value = item.values()[0], item.values()[1]
                    data['values'][ds.component][ds.datasource].append(
                        (value, calendar.timegm(timestamp.timetuple()))
                    )

        defer.returnValue(data)
