######################################################################
#
# Copyright (C) Zenoss, Inc. 2013-2016, all rights reserved.
#
# This content is made available according to terms specified in
# License.zenoss under the directory where your Zenoss product is
# installed.
#
######################################################################

import json
import re
from logging import getLogger
log = getLogger('zen.python')

import boto
import boto.ec2
import boto.sqs
import boto.s3
import boto.vpc

from boto.sqs.message import RawMessage
from twisted.internet import defer, threads

from Products.DataCollector.plugins.DataMaps import ObjectMap, RelationshipMap
from Products.ZenEvents import ZenEventClasses
from Products.ZenUtils.Utils import prepId
from ZenPacks.zenoss.PythonCollector.datasources.PythonDataSource \
    import PythonDataSourcePlugin

from ZenPacks.zenoss.AWS import MODULE_NAME
from ZenPacks.zenoss.AWS.modeler.plugins.aws.EC2 import (
    INSTANCE_FILTERS, get_instance_data, check_tag, path_to_pem,
)
from ZenPacks.zenoss.AWS.txboto import (
    TwistedEC2Connection, TwistedVPCConnection, TwistedS3Connection,
)
from ZenPacks.zenoss.AWS.utils import prodState


class AWSBasePlugin(PythonDataSourcePlugin):
    """
    Subclass of PythonDataSourcePlugin to monitor AWS components.
    """
    proxy_attributes = (
        'ec2accesskey', 'ec2secretkey',
    )
    component = None

    @classmethod
    def params(cls, datasource, context):
        try:
            region = datasource.talesEval(datasource.region, context)
        except:
            return {}
        return {
            'region': region,
        }

    def onSuccess(self, result, config):
        for component in result["values"].keys():
            result['events'].insert(0, {
                'component': component,
                'summary': 'Monitoring ok',
                'eventClass': '/Status',
                'eventKey': 'aws_result',
                'severity': ZenEventClasses.Clear,
            })
        return result

    def onError(self, result, config):
        data = self.new_data()
        str_res = str(result)

        eventKey = 'aws_result'
        if not self.component:
            eventKey = 'aws_bucket_result'
        if "<Message>" in str_res and self.component:
            m = re.search('<Message>(.+?)</Message>', str_res)
            if m:
                res = m.group(1)
                log.info(res)
                if self.component in str_res:
                    data['events'].append({
                        'component': self.component,
                        'summary': res,
                        'eventClass': '/Status',
                        'eventKey': 'aws_result',
                        'severity': ZenEventClasses.Info,
                    })
                    return data
        else:
            if 'IndexError' in str(result.type):
                summary = 'The component {0} does not exist.'.format(
                    self.component
                )
                severity = ZenEventClasses.Info
            elif ('timed out' in str_res) or ("name resolution" in str_res) or\
                 ("service not known" in str_res) or\
                 ("argument of type 'int' is not iterable" in str_res):
                summary = "Connection timed out or network problems."
                severity = ZenEventClasses.Error
            else:
                summary = str_res
                severity = ZenEventClasses.Error

            data['events'].append({
                'component': self.component,
                'summary': summary,
                'eventClass': '/Status',
                'eventKey': eventKey,
                'severity': severity,
            })
            return data


class S3BucketPlugin(AWSBasePlugin):
    """
    Subclass of AWSBasePlugin to monitor AWS S3Buckets.
    """

    DEFAULT_REGION = 'us-east-1'

    @defer.inlineCallbacks
    def collect(self, config):
        data = self.new_data()
        bucket_keys = {}
        # Add support Signature V4 for 's3' into boto config
        if not boto.config.get('s3', 'use-sigv4'):
            boto.config.add_section('s3')
            boto.config.set('s3', 'use-sigv4', 'True')

        ds0 = config.datasources[0]

        s3connection = TwistedS3Connection.adapt(boto.s3.connect_to_region(
            S3BucketPlugin.DEFAULT_REGION,
            aws_access_key_id=ds0.ec2accesskey,
            aws_secret_access_key=ds0.ec2secretkey,
        ))

        try:
            buckets = yield s3connection.get_all_buckets()
        except Exception as ex:
            log.error("Unable to get buckets for %s region, %s",
                      S3BucketPlugin.DEFAULT_REGION, str(ex))
        else:
            data['events'].append({
                'summary': 'Monitoring ok',
                'eventClass': '/Status',
                'eventKey': 'aws_bucket_result',
                'severity': ZenEventClasses.Clear,
            })

            bucket_region = {}
            connections = {}
            for ds in config.datasources:
                region = ds.params['region']
                if region == "" or region is None:
                    continue

                bucket_region[ds.component] = region
                if region not in connections:
                    s3conn = boto.s3.connect_to_region(
                        region,
                        aws_access_key_id=ds0.ec2accesskey,
                        aws_secret_access_key=ds0.ec2secretkey,)
                    if s3conn:
                        connections[region] = TwistedS3Connection.adapt(s3conn)
                    else:
                        log.error("Unable to initiate an S3 connection to "
                                  "region: %s, plugin: %s, template: %s, "
                                  "datasource: %s, device: %s, component: %s",
                                  region, ds.plugin_classname, ds.template,
                                  ds.datasource, ds.device, ds.component)

            deferreds = []
            names = []
            for bucket in buckets:

                region = bucket_region.get(bucket.name)
                if region is None:
                    log.error("Bucket %s: Unable to determine region: skipping collection until this S3Bucket is modeled", bucket.name)
                    continue

                if bool(connections):
                    if connections.get(region) is None:
                        log.error("Bucket %s: Unable to connect to region %s: skipping collection", bucket.name, region)
                        continue

                    # to retrieve further information about the bucket, we need to
                    # query the right region.   Switch its connection over to the
                    # right one.
                    bucket.connection = connections[region]

                d = bucket.get_all_keys()
                names.append(bucket.name)
                deferreds.append(d)

            # Collect all of the bucket keys in parallel
            results = yield defer.DeferredList(deferreds, consumeErrors=True)

            for i, value in enumerate(results):
                success, result = value
                name = names[i]

                if success:
                    # value is a list of keys.
                    bucket_keys.update({name: result})
                else:
                    # result is a twisted Failure object, and result.value is the
                    # exception that occurred.
                    bucket_keys.update({name: result.value})

            for ds in config.datasources:
                self.component = ds.component
                keys = bucket_keys.get(ds.component) or []
                if not isinstance(keys, list):
                    data['events'].append({
                        'component': self.component,
                        'summary': str(keys),
                        'eventClass': '/Status',
                        'eventKey': 'aws_result',
                        'severity': ZenEventClasses.Error,
                    })
                    continue
                data['values'][ds.component] = dict(
                    keys_count=(len(keys), 'N'),
                    total_size=(sum([key.size for key in keys]), 'N'),
                )

        defer.returnValue(data)


@defer.inlineCallbacks
def get_instances_relmap(region_id, ds, instances, images,
                         instance_states, region_conn):
    """
    Return instances RelationshipMap given region_id and an InstanceInfo
    ResultSet.
    """
    image_filters = [i.image_id for i in instances]
    if image_filters:
        try:
            image_data = yield region_conn.get_all_images(
                image_ids=image_filters)
        except Exception as e:
            log.warning("Unable to retrieve image data for "
                        "plugin: %s, template: %s, datasource: %s, "
                        "device: %s, component: %s, %s",
                        ds.plugin_classname, ds.template, ds.datasource,
                        ds.device, ds.component, str(e))
        else:
            images.extend(image_data)

    instance_data = []
    image_ids = [x.id for x in images]
    for instance in instances:
        data = get_instance_data(instance, image_ids)
        data.update({
            'guest': check_tag(ds.zAWSDiscover, instance.tags),
            'pem_path': path_to_pem(region_id, ds.zAWSRegionPEM),
        })
        instance_states[data['id']] = prodState(data['state'])
        instance_data.append(data)

    defer.returnValue(
        RelationshipMap(
            compname='regions/%s' % region_id,
            relname='instances',
            modname=MODULE_NAME['EC2Instance'],
            objmaps=instance_data
        )
    )


class EC2RegionPlugin(AWSBasePlugin):
    """
    Retrieves information about soft limits for region
    and updates statuses of EC2 instances and volumes.
    """

    proxy_attributes = (
        'ec2accesskey',
        'ec2secretkey',
        'zAWSDiscover',
        'zAWSRegionPEM',
        'zAWSRemodelEnabled',
        'getDiscoverGuests',
    )

    @defer.inlineCallbacks
    def collect(self, config):
        data = self.new_data()
        instance_states = {}
        for ds in config.datasources:
            region_id = self.component = ds.component
            ec2regionconn = TwistedEC2Connection.adapt(boto.ec2.connect_to_region(
                region_id,
                aws_access_key_id=ds.ec2accesskey,
                aws_secret_access_key=ds.ec2secretkey,
            ))

            vpcregionconn = TwistedVPCConnection.adapt(boto.vpc.connect_to_region(
                region_id,
                aws_access_key_id=ds.ec2accesskey,
                aws_secret_access_key=ds.ec2secretkey,
            ))

            instances_d = ec2regionconn.get_only_instances(
                filters=INSTANCE_FILTERS
            )
            volumes_d = ec2regionconn.get_all_volumes()
            elastic_ips_d = ec2regionconn.get_all_addresses()
            subnets_d = vpcregionconn.get_all_subnets()
            groups_d = ec2regionconn.get_all_security_groups()

            # Retrieve this data in parallel (per region)
            deferreds = [instances_d, volumes_d, elastic_ips_d, subnets_d, groups_d]
            labels = ('Instances', 'Volumes', 'Elastic IPs', 'Subnets', 'Security Groups')
            results = yield defer.DeferredList(deferreds, consumeErrors=True)

            values = {}
            instances = volumes = []
            for i in range(len(results)):
                success, result = results[i]

                if not success:
                    log.error("Error retrieving %s for %s: %s",
                              labels[i], region_id, result.getErrorMessage())

                    # Give up on this region, since some of the data
                    # couldn't be retrieved successfully.
                    continue

                if i == 0:
                    values['instances_count'] = (len(result), 'N')
                    instances = result
                elif i == 1:
                    values['volumes_count'] = (len(result), 'N')
                    volumes = result
                elif i == 2:
                    values['elastic_ips_count'] = (len(result), 'N')
                elif i == 3:
                    values['subnets_count'] = (len(result), 'N')
                elif i == 4:
                    values['vpc_security_groups_count'] = (len(result), 'N')
                    rules_count = 0
                    for group in result:
                        rules_count = max(len(group.rules), rules_count)
                    values['vpc_security_rules_count'] = (rules_count, 'N')
                else:
                    raise ValueError("Error retrieving %s for %s: %s" %
                                     (labels[i], region_id, result.getErrorMessage()))

            data['values'][region_id] = values

            # Previously zAWSRemodelEnabled was a string, fix for this case
            if str(ds.zAWSRemodelEnabled).lower() == 'true':
                instances_relmap = yield get_instances_relmap(
                    region_id, ds, instances, [], instance_states, ec2regionconn)
                data['maps'].append(instances_relmap)
            else:
                # InstanceState moved here for optimization
                ins_res = {}
                # update instance map only when there are new instances
                if len(instances) > 0:
                    for instance in instances:
                        ins_res[prepId(instance.id)] = instance.state
                    data['maps'].append(ObjectMap({
                        "compname": "regions/%s" % region_id,
                        "modname": "Instances states",
                        "setInstancesStates": ins_res
                    }))

            # VolumeState moved here for optimization
            vol_res = {}
            # update volume map only when there are new volumes
            if len(volumes) > 0:
                for volume in volumes:
                    vol_res[prepId(volume.id)] = volume.status
                data['maps'].append(ObjectMap({
                    "compname": "regions/%s" % region_id,
                    "modname": "Volumes states",
                    "setVolumesStatuses": vol_res
                }))

        ds0 = config.datasources[0]
        if str(ds0.zAWSRemodelEnabled).lower() == 'true':
            # Run "setDiscoverGuests" every time during the monitoring
            data['maps'].append(ObjectMap({
                "modname": "Guest update",
                "setDiscoverGuests": False if ds0.getDiscoverGuests
                else True,
            }))
        defer.returnValue(data)


def get_messages(queue):
    messages = {}
    message_count = -1
    while message_count < len(messages):
        message_count = len(messages)
        res = queue.get_messages(num_messages=10, visibility_timeout=3)
        for message in res:
            messages[message.id] = message._body
    return messages


class SQSQueuePlugin(AWSBasePlugin):
    """
    Subclass of AWSBasePlugin to monitor AWS SQSQueue.
    """

    def collect(self, config):
        def inner():
            data = self.new_data()
            for ds in config.datasources:
                self.component = ds.component
                name = ds.component[len('queue_'):]
                region = ds.params['region']
                sqsconnection = boto.sqs.connect_to_region(
                    region,
                    aws_access_key_id=ds.ec2accesskey,
                    aws_secret_access_key=ds.ec2secretkey,
                )
                queue = sqsconnection.get_queue(name)
                if queue:
                    queue.set_message_class(RawMessage)
                    messages = get_messages(queue)

                    data['events'].append({
                        'component': self.component,
                        'summary': "Monitoring ok",
                        'eventClass': '/Status',
                        'eventKey': 'aws_result',
                        'severity': ZenEventClasses.Clear,
                    })

                    for id, text in messages.iteritems():
                        try:
                            msg_dict = json.loads(text)
                        except ValueError:
                            message = summary = text
                        else:
                            message = msg_dict.get('Message')
                            summary = msg_dict.get('Subject')

                        data['events'].append({
                            'message': message,
                            'summary': summary,
                            'device': config.id,
                            'component': self.component,
                            'eventKey': id,
                            'severity': ZenEventClasses.Info,
                            'eventClass': '/AWS/SQSMessage',
                        })
                else:
                    data['events'].append({
                        'summary': 'Queue "%s" does not exist' % name,
                        'device': config.id,
                        'component': self.component,
                        'severity': ZenEventClasses.Info,
                        'eventClass': '/Status',
                    })
            return data

        return threads.deferToThread(inner)


class ZonePlugin(AWSBasePlugin):
    """
    Subclass of AWSBasePlugin to monitor AWS Zone.
    """

    @defer.inlineCallbacks
    def collect(self, config):
        data = self.new_data()
        for ds in config.datasources:
            self.component = ds.component
            region = ds.params['region']
            ec2regionconn = TwistedEC2Connection.adapt(boto.ec2.connect_to_region(
                region,
                aws_access_key_id=ds.ec2accesskey,
                aws_secret_access_key=ds.ec2secretkey
            ))

            try:
                zones = yield ec2regionconn.get_all_zones(ds.component)
            except Exception as ex:
                log.error("Unable to get zones for %s region, "
                          "plugin: %s, template: %s, datasource: %s, "
                          "device: %s, component: %s, %s",
                          region, ds.plugin_classname, ds.template,
                          ds.datasource, ds.device, ds.component, str(ex))
            else:
                zone = zones.pop()
                if zone.state == 'available':
                    severity = ZenEventClasses.Clear
                    data['events'].append({
                        'summary': "Monitoring Ok",
                        'component': ds.component,
                        'severity': severity,
                        'eventKey': 'aws_result',
                        'eventClass': '/Status',
                    })
                else:
                    severity = ZenEventClasses.Warning
                data['events'].append({
                    'summary': 'Zone state is {0}'.format(zone.state),
                    'component': ds.component,
                    'eventKey': 'ZoneStatus',
                    'severity': severity,
                    'eventClass': '/Status',
                })

                data['maps'].append(ObjectMap({
                    "compname": "regions/%s/zones/%s" % (
                        region, ds.component),
                    "modname": "Zone state",
                    "state": zone.state
                }))

        defer.returnValue(data)


class EC2VPCSubnetPlugin(AWSBasePlugin):
    """
    Subclass of AWSBasePlugin to monitor AWS VPC Subnets.
    """

    @defer.inlineCallbacks
    def collect(self, config):
        data = self.new_data()
        for ds in config.datasources:
            self.component = ds.component
            region = ds.params['region']
            vpcregionconn = TwistedVPCConnection.adapt(boto.vpc.connect_to_region(
                region,
                aws_access_key_id=ds.ec2accesskey,
                aws_secret_access_key=ds.ec2secretkey,
            ))
            try:
                subnets = yield vpcregionconn.get_all_subnets(ds.component)
            except Exception as ex:
                log.error("Unable to get zones for %s region, "
                          "plugin: %s, template: %s, datasource: %s, "
                          "device: %s, component: %s, %s",
                          region, ds.plugin_classname, ds.template,
                          ds.datasource, ds.device, ds.component, str(ex))
            else:
                subnet = subnets.pop()

                data['values'][ds.component] = dict(
                    available_ip_address_count=(
                        subnet.available_ip_address_count, 'N'
                    )
                )

                data['maps'].append(ObjectMap({
                    "compname": "regions/%s/vpc_subnets/%s" % (
                        region, ds.component),
                    "modname": "Subnet state",
                    "state": subnet.state
                }))

        defer.returnValue(data)


CONNECTION_TYPE = {
    'EC2VPC': 'vpc',
    'EC2Snapshot': 'ec2',
    'EC2Volume': 'ec2',
    'VPNGateway': 'vpc',
    'EC2Image': 'ec2',
    'EC2Instance': 'ec2',
    'EC2Instance-Detailed': 'ec2',
    # 'EC2Instance-Custom': 'ec2',
}


# Plugins for components' state remodeling.
class EC2BaseStatePlugin(AWSBasePlugin):
    """
    Subclass of AWSBasePlugin to monitor AWS components' states.
    """
    ec2regionconn = None
    vpcregionconn = None

    def connect_to_region(self, ds):
        region = ds.params['region']

        if CONNECTION_TYPE.get(ds.template) == 'ec2':
            self.ec2regionconn = TwistedEC2Connection.adapt(boto.ec2.connect_to_region(
                region,
                aws_access_key_id=ds.ec2accesskey,
                aws_secret_access_key=ds.ec2secretkey,
            ))
        else:
            self.vpcregionconn = TwistedVPCConnection.adapt(boto.vpc.connect_to_region(
                region,
                aws_access_key_id=ds.ec2accesskey,
                aws_secret_access_key=ds.ec2secretkey,
            ))

    @defer.inlineCallbacks
    def results_to_maps(self, region, component):
        """Return Object map for the component status remodeling.  """

    def gen_events(self, ds, data):
        """ Generate the events.  """

    @defer.inlineCallbacks
    def collect(self, config):
        data = self.new_data()
        for ds in config.datasources:
            self.component = ds.component
            region = ds.params['region']
            self.connect_to_region(ds)

            try:
                maps = yield self.results_to_maps(region, ds.component)
            except Exception as ex:
                log.error("Unable to get maps for %s region, "
                          "plugin: %s, template: %s, datasource: %s, "
                          "device: %s, component: %s, %s",
                          region, ds.plugin_classname, ds.template,
                          ds.datasource, ds.device, ds.component, str(ex))
            else:
                if maps:
                    data['maps'].append(maps)

                self.gen_events(ds, data)

        defer.returnValue(data)


class EC2VPCStatePlugin(EC2BaseStatePlugin):
    """
    Subclass of EC2BaseStatePlugin to monitor AWS VPC state.
    """

    @defer.inlineCallbacks
    def results_to_maps(self, region, component):
        vpcs = yield self.vpcregionconn.get_all_vpcs(component)
        vpc = vpcs.pop()
        defer.returnValue(ObjectMap({
            "compname": "regions/%s/vpcs/%s" % (
                region, component),
            "modname": "VPC state",
            "state": vpc.state
        }))


class VPNGatewayStatePlugin(EC2BaseStatePlugin):
    """
    Subclass of EC2BaseStatePlugin to monitor AWS Gateways state.
    """

    @defer.inlineCallbacks
    def results_to_maps(self, region, component):
        gateways = yield self.vpcregionconn.get_all_vpn_gateways(component)
        gateway = gateways.pop()
        defer.returnValue(ObjectMap({
            "compname": "regions/%s/vpn_gateways/%s" % (
                region, component),
            "modname": "VPNGateway state",
            "state": gateway.state
        }))


class EC2UnreservedInstancesPlugin(AWSBasePlugin):
    """
    Subclass of EC2BaseStatePlugin to check if instance could be reserved
    """

    @defer.inlineCallbacks
    def unreserved_instance_count(self, ec2_conn, instance):
        if instance.state != 'running':
            defer.returnValue(0)  # stoppend instances not need discount
            return
        if instance.spot_instance_request_id:
            defer.returnValue(0)  # spot instances cannot use discount
            return

        # getting running instances of the same type
        # in the same availability zone
        instances = yield ec2_conn.get_only_instances(filters={
            'instance-state-name': 'running',
            'instance-type': instance.instance_type,
            'availability-zone': instance.placement,
        })
        # and not spot instances
        instances = filter(lambda i: not i.spot_instance_request_id, instances)

        # getting active reserved instances of the same type
        # in the same availability zone
        reserved = yield ec2_conn.get_all_reserved_instances(filters={
            'state': 'active',
            'instance-type': instance.instance_type,
            'availability-zone': instance.placement,
        })
        defer.returnValue(len(instances) - len(reserved))

    @defer.inlineCallbacks
    def collect(self, config):
        data = self.new_data()
        for ds in config.datasources:
            self.component = ds.component
            region = ds.params['region']
            ec2_conn = TwistedEC2Connection.adapt(boto.ec2.connect_to_region(
                region,
                aws_access_key_id=ds.ec2accesskey,
                aws_secret_access_key=ds.ec2secretkey,
            ))
            instances = yield ec2_conn.get_only_instances(ds.component)
            instance = instances.pop()
            c = yield self.unreserved_instance_count(ec2_conn, instance)

            event = None
            if c == 1:
                event = 'This instance could be reserved'
            elif c > 1:
                event = 'There are {0} instances of this type in this ' \
                    'availability zone which could be reserved'.format(c)

            if event:
                data['events'].append({
                    'summary': event,
                    'device': config.id,
                    'component': ds.component,
                    'severity': ZenEventClasses.Info,
                    'eventClass': '/AWS/Suggestion',
                })

        defer.returnValue(data)


class EC2UnusedReservedInstancesPlugin(AWSBasePlugin):

    @defer.inlineCallbacks
    def unused_reserved_instances_count(self, ec2_conn, reserved_instance):
        # getting running instances of the same type
        # in the same availability zone
        instances = yield ec2_conn.get_only_instances(filters={
            'instance-state-name': 'running',
            'instance-type': reserved_instance.instance_type,
            'availability-zone': reserved_instance.availability_zone,
        })
        # and not spot instances
        instances = filter(lambda i: not i.spot_instance_request_id, instances)

        # getting active reserved instances of the same type
        # in the same availability zone
        reserved = yield ec2_conn.get_all_reserved_instances(filters={
            'state': 'active',
            'instance-type': reserved_instance.instance_type,
            'availability-zone': reserved_instance.availability_zone,
        })
        defer.returnValue(len(reserved) - len(instances))

    @defer.inlineCallbacks
    def collect(self, config):
        data = self.new_data()
        for ds in config.datasources:
            self.component = ds.component
            region = ds.params['region']
            ec2_conn = TwistedEC2Connection.adapt(boto.ec2.connect_to_region(
                region,
                aws_access_key_id=ds.ec2accesskey,
                aws_secret_access_key=ds.ec2secretkey,
            ))

            reserved_instances = yield ec2_conn.get_all_reserved_instances(
                ds.component
            )
            reserved_instance = reserved_instances.pop()

            c = yield self.unused_reserved_instances_count(
                ec2_conn,
                reserved_instance
            )

            event = None
            if c == 1:
                event = 'This reserved instance is unused'
            elif c > 1:
                event = 'There are {0} unused reserved instances of ' \
                    'this type in this availability zone'.format(c)

            if event:
                data['events'].append({
                    'summary': event,
                    'device': config.id,
                    'component': ds.component,
                    'severity': ZenEventClasses.Info,
                    'eventClass': '/AWS/Suggestion',
                })

        defer.returnValue(data)
