##############################################################################
#
# Copyright (C) Zenoss, Inc. 2016, all rights reserved.
#
# This content is made available according to terms specified in
# License.zenoss under the directory where your Zenoss product is installed.
#
##############################################################################


import logging
log = logging.getLogger('zen.txboto')

from twisted.internet import defer
from twisted.internet.error import ConnectionClosed, TimeoutError
from twisted.web.error import Error as HTTPError

from boto.ec2.connection import EC2Connection
from boto.ec2.cloudwatch import CloudWatchConnection
from boto.exception import S3ResponseError
from boto.vpc import VPCConnection
from boto.s3.connection import S3Connection
from boto.s3.key import Key
from boto.s3.bucket import Bucket

import boto.handler
from boto.resultset import ResultSet
import six
import xml.sax
from urllib import urlencode

from ZenPacks.zenoss.AWS.utils import getPage

# this can be a bit noisy if enabled.
# logging.getLogger('boto').setLevel(log.getEffectiveLevel())


# Global per-API-endpoint/user concurrent connection limit
MAX_PARALLEL = 128


class TwistedConnectionBase(object):

    def concurrent_key(self, http_request):
        return (http_request.host, self.provider.get_access_key())

    @defer.inlineCallbacks
    def _execute_http_request(self, http_request, action=None):
        if http_request.method == 'POST':
            http_request.body = urlencode(sorted(http_request.params.iteritems()))

        if action:
            http_request.params['Action'] = action
        if getattr(self, 'APIVersion', None):
            http_request.params['Version'] = self.APIVersion

        # Stringify all the headers, or twisted's web client will be cranky.
        http_request.headers = {k: str(v) for k, v in http_request.headers.iteritems()}

        # Sign the request
        log.debug('Token: %s' % self.provider.security_token)
        http_request.authorize(connection=self)

        # Only force header for non-s3 connections, because s3 uses
        # an older signing method + bucket resource URLs that include
        # the port info. All others should be now be up to date and
        # not include the port.
        if 's3' not in self._required_auth_capability():
            if not getattr(self, 'anon', False):
                self.set_host_header(http_request)

        # Stringify all the headers, or twisted's web client will be cranky.
        http_request.headers = {k: str(v) for k, v in http_request.headers.iteritems()}

        url = str('%s://%s:%s%s' % (http_request.protocol, http_request.host, http_request.port, http_request.path))

        log.debug('Method: %s' % http_request.method)
        log.debug('URL: %s' % url)
        log.debug('Data: %s' % http_request.body)
        log.debug('Headers: %s' % http_request.headers)
        log.debug('Params: %s' % http_request.params)

        if http_request.method == 'POST':
            response = yield getPage(
                concurrent_key=self.concurrent_key(http_request),
                concurrent_limit=MAX_PARALLEL,
                method=http_request.method,
                url=url,
                headers=http_request.headers,
                postdata=http_request.body,
                agent=http_request.headers['User-Agent']
            )
            defer.returnValue(response)
        elif http_request.method == 'GET':
            response = yield getPage(
                concurrent_key=self.concurrent_key(http_request),
                concurrent_limit=MAX_PARALLEL,
                method=http_request.method,
                url=url,
                headers=http_request.headers,
                agent=http_request.headers['User-Agent']
            )
            defer.returnValue(response)
        else:
            raise Exception("http_request method %s not implemented" % http_request.method)


class TwistedEC2Connection(EC2Connection, TwistedConnectionBase):

    @staticmethod
    def adapt(ec2connection):
        if type(ec2connection) != EC2Connection:
            raise Exception("TwistedEC2Connection() must be supplied with an EC2Connection (not %s)" % ec2connection)
        ec2connection.__class__ = TwistedEC2Connection
        return ec2connection

    @defer.inlineCallbacks
    def make_request(self, action, params=None, path='/', verb='GET'):
        http_request = self.build_base_http_request(verb, path, None,
                                                    params, {}, '',
                                                    self.host)

        result = yield self._execute_http_request(http_request, action)
        defer.returnValue(result)

    @defer.inlineCallbacks
    def get_list(self, action, params, markers, path='/',
                 parent=None, verb='GET'):
        if not parent:
            parent = self

        try:
            body = yield self.make_request(action, params, path, verb)
        except HTTPError, error:
            log.debug('%s %s' % (error.status, error.message))
            # zen-24209. QA wants an ERROR log entry for
            # 'RequestTimeTooSkewed', instead of a DEBUG one
            if 'RequestTimeTooSkewed' in error.response:
                log.error('%s. Please make sure your Zenoss system time is accurate, and/or ntpd is running.' % error.response)
            else:
                log.debug('%s' % error.response)
            raise self.ResponseError(error.status, error.message, error.response)
        except (ConnectionClosed, TimeoutError) as error:
            raise self.ResponseError('', str(error), '')

        log.debug(body)

        if not body:
            log.debug('Null body %s' % body)
            raise self.ResponseError(200, 'OK', body)

        rs = ResultSet(markers)
        h = boto.handler.XmlHandler(rs, parent)
        if isinstance(body, six.text_type):
            body = body.encode('utf-8')
        xml.sax.parseString(body, h)
        defer.returnValue(rs)

    @defer.inlineCallbacks
    def get_object(self, action, params, cls, path='/',
                   parent=None, verb='GET'):
        if not parent:
            parent = self

        try:
            body = yield self.make_request(action, params, path, verb)
        except HTTPError, error:
            log.debug('%s %s' % (error.status, error.message))
            if 'RequestTimeTooSkewed' in error.response:
                log.error('%s. Please make sure your Zenoss system time is accurate, and/or ntpd is running.' % error.response)
            else:
                log.debug('%s' % error.response)
            raise self.ResponseError(error.status, error.message, error.response)

        log.debug(body)

        if not body:
            log.debug('Null body %s' % body)
            raise self.ResponseError(200, 'OK', body)

        obj = cls(parent)
        h = boto.handler.XmlHandler(obj, parent)
        if isinstance(body, six.text_type):
            body = body.encode('utf-8')
        xml.sax.parseString(body, h)
        defer.returnValue(obj)

    @defer.inlineCallbacks
    def get_status(self, action, params, path='/', parent=None, verb='GET'):
        if not parent:
            parent = self

        try:
            body = yield self.make_request(action, params, path, verb)
        except HTTPError, error:
            log.debug('%s %s' % (error.status, error.message))
            if 'RequestTimeTooSkewed' in error.response:
                log.error('%s. Please make sure your Zenoss system time is accurate, and/or ntpd is running.' % error.response)
            else:
                log.debug('%s' % error.response)
            raise self.ResponseError(error.status, error.message, error.response)

        log.debug(body)

        if not body:
            log.debug('Null body %s' % body)
            raise self.ResponseError(200, 'OK', body)

        rs = ResultSet()
        h = boto.handler.XmlHandler(rs, parent)
        xml.sax.parseString(body, h)
        defer.returnValue(rs.status)

    # Any method that only calls get_status, get_object, or get_list,
    # and returns the value it returns (now a deferred) will work without
    # further modification.

    # Methods that call other methods, though, will need to be subclassed to
    # yield in the appropriate spots:

    @defer.inlineCallbacks
    def get_only_instances(self, instance_ids=None, filters=None,
                           dry_run=False, max_results=None):
        next_token = None
        retval = []
        while True:
            reservations = yield self.get_all_reservations(instance_ids=instance_ids,
                                                           filters=filters,
                                                           dry_run=dry_run,
                                                           max_results=max_results,
                                                           next_token=next_token)
            retval.extend([instance for reservation in reservations for
                           instance in reservation.instances])
            next_token = reservations.next_token
            if not next_token:
                break

        defer.returnValue(retval)


class TwistedCloudWatchConnection(CloudWatchConnection, TwistedEC2Connection):

    @staticmethod
    def adapt(cloudwatchconnection):
        if type(cloudwatchconnection) != CloudWatchConnection:
            raise Exception("TwistedCloudWatchConnection() must be supplied with an CloudWatchConnection (not %s)" % cloudwatchconnection)
        cloudwatchconnection.__class__ = TwistedCloudWatchConnection
        return cloudwatchconnection


class TwistedVPCConnection(TwistedEC2Connection, VPCConnection, TwistedConnectionBase):

    @staticmethod
    def adapt(vpcconnection):
        if type(vpcconnection) != VPCConnection:
            raise Exception("TwistedVPCConnection() must be supplied with an VPCConnection (not %s)" % vpcconnection)
        vpcconnection.__class__ = TwistedVPCConnection
        return vpcconnection


class TwistedS3Connection(S3Connection, TwistedConnectionBase):

    @staticmethod
    def adapt(s3connection):
        if type(s3connection) != S3Connection:
            raise Exception("TwistedS3Connection() must be supplied with an S3Connection (not %s)" % s3connection)
        s3connection.__class__ = TwistedS3Connection
        s3connection.set_bucket_class(TwistedS3Bucket)

        return s3connection

    def concurrent_key(self, http_request):
        # Note: self.host is the s3 endpoint, and http_request.host may
        # include specific bucket info.  We use the more general one to
        # throttle per-region, not per bucket.
        return (self.host, self.provider.get_access_key())

    @defer.inlineCallbacks
    def make_request(self, method, bucket='', key='', headers=None, data='',
                     query_args=None, sender=None, override_num_retries=None,
                     retry_handler=None):
        if isinstance(bucket, self.bucket_class):
            bucket = bucket.name
        if isinstance(key, Key):
            key = key.name
        path = self.calling_format.build_path_base(bucket, key)
        boto.log.debug('path=%s' % path)
        auth_path = self.calling_format.build_auth_path(bucket, key)
        boto.log.debug('auth_path=%s' % auth_path)
        host = self.calling_format.build_host(self.server_name(), bucket)
        if query_args:
            path += '?' + query_args
            boto.log.debug('path=%s' % path)
            auth_path += '?' + query_args
            boto.log.debug('auth_path=%s' % auth_path)

        params = {}
        http_request = self.build_base_http_request(method, path, auth_path,
                                                    params, headers, data, host)

        result = yield self._execute_http_request(http_request)
        defer.returnValue(result)

    @defer.inlineCallbacks
    def get_all_buckets(self, headers=None):
        try:
            body = yield self.make_request('GET', headers=headers)
        except HTTPError, error:
            log.debug('%s %s' % (error.status, error.message))
            if 'RequestTimeTooSkewed' in error.response:
                log.error('%s. Please make sure your Zenoss system time is accurate, and/or ntpd is running.' % error.response)
            else:
                log.debug('%s' % error.response)
            raise self.provider.storage_response_error(
                error.status, error.message, error.response)
        except (ConnectionClosed, TimeoutError) as error:
            raise S3ResponseError('', str(error), '')

        if not body:
            log.debug('Null body %s' % body)
            raise S3ResponseError(200, 'OK', body)

        rs = ResultSet([('Bucket', self.bucket_class)])
        h = boto.handler.XmlHandler(rs, self)
        if not isinstance(body, bytes):
            body = body.encode('utf-8')
        xml.sax.parseString(body, h)
        defer.returnValue(rs)


class TwistedS3Bucket(Bucket):

    @defer.inlineCallbacks
    def _get_all(self, element_map, initial_query_string='',
                 headers=None, **params):
        query_args = self._get_all_query_args(
            params,
            initial_query_string=initial_query_string
        )
        try:
            body = yield self.connection.make_request('GET', self.name,
                                                      headers=headers,
                                                      query_args=query_args)
        except HTTPError, error:
            log.debug('%s %s' % (error.status, error.message))
            if 'RequestTimeTooSkewed' in error.response:
                log.error('%s. Please make sure your Zenoss system time is accurate, and/or ntpd is running.' % error.response)
            else:
                log.debug('%s' % error.response)
            raise self.connection.provider.storage_response_error(
                error.status, error.message, error.response)

        if not body:
            log.debug('Null body %s' % body)
            raise self.ResponseError(200, 'OK', body)

        rs = ResultSet(element_map)
        h = boto.handler.XmlHandler(rs, self)
        if not isinstance(body, bytes):
            body = body.encode('utf-8')
        xml.sax.parseString(body, h)
        defer.returnValue(rs)

    @defer.inlineCallbacks
    def get_location(self):
        """
        Returns the LocationConstraint for the bucket.

        :rtype: str
        :return: The LocationConstraint for the bucket or the empty
            string if no constraint was specified when bucket was created.
        """
        try:
            body = yield self.connection.make_request('GET', self.name,
                                                      query_args='location')
        except HTTPError, error:
            log.debug('%s %s' % (error.status, error.message))
            if 'RequestTimeTooSkewed' in error.response:
                log.error('%s. Please make sure your Zenoss system time is accurate, and/or ntpd is running.' % error.response)
            else:
                log.debug('%s' % error.response)
            raise self.connection.provider.storage_response_error(
                error.status, error.message, error.response)

        if not body:
            log.debug('Null body %s' % body)
            raise self.ResponseError(200, 'OK', body)

        rs = ResultSet(self)
        h = boto.handler.XmlHandler(rs, self)
        if not isinstance(body, bytes):
            body = body.encode('utf-8')
        xml.sax.parseString(body, h)
        defer.returnValue(rs.LocationConstraint)
