##############################################################################
#
# Copyright (C) Zenoss, Inc. 2013-2016, all rights reserved.
#
# This content is made available according to terms specified in
# License.zenoss under the directory where your Zenoss product is installed.
#
##############################################################################

import logging
log = logging.getLogger('zen.AWS')

import base64
from cStringIO import StringIO
import hashlib
import hmac
import datetime
from lxml import etree
import os
import random
import urllib
from urlparse import urlunparse

from twisted.internet import defer, reactor
from twisted.internet.error import (
    ConnectionRefusedError, DNSLookupError, TimeoutError, ConnectionClosed
)
from twisted.web import http
from twisted.web import client as txwebclient

from zope.event import notify

from Products.AdvancedQuery import Eq, Or
from Products.ZenUtils.Utils import prepId
from Products.Zuul.catalog.events import IndexingEvent
from Products.Zuul.interfaces import ICatalogTool
from Products.ZenUtils import Map

SEMAPHORES = {}
DNS_CACHE_TIME = 300


def addLocalLibPath():
    """
    Helper to add the ZenPack's lib directory to PYTHONPATH.
    """
    import os
    import site

    site.addsitedir(os.path.join(os.path.dirname(__file__), 'lib'))


class ProxyWebClient(object):
    """web methods with proxy support."""

    dns_cache = Map.Timed({}, DNS_CACHE_TIME)

    def __init__(self, url, username=None, password=None):
        # get scheme used by url
        scheme, host, port, path = twisted_web_client_parse(url)
        envname = '%s_proxy' % scheme
        self.use_proxy = False
        self.proxy_host = None
        self.proxy_port = None
        if envname in os.environ.keys():
            proxy = os.environ.get('%s_proxy' % scheme)
            if proxy:
                # using proxy server
                # host:port identifies a proxy server
                # url is the actual target
                self.use_proxy = True
                scheme, host, port, path = twisted_web_client_parse(proxy)
                self.proxy_host = host
                self.proxy_port = port
                self.username = username
                self.password = password
        else:
            self.host = host
            self.port = port
        self.path = url
        self.url = url

    @defer.inlineCallbacks
    def resolve(self, hostname):
        """Asynchronously resolve a hostname, with caching."""
        ip = self.dns_cache.get(hostname, None)
        if ip is None:
            ip = yield reactor.resolve(hostname)
            self.dns_cache[hostname] = ip

        defer.returnValue(ip)

    @defer.inlineCallbacks
    def get_page(self, contextFactory=None, description=None, *args, **kwargs):
        if description is None:
            description = self.url

        log.debug("Requesting %s", description)

        scheme, _, _, _ = twisted_web_client_parse(self.url)
        factory = txwebclient.HTTPClientFactory(self.url, *args, **kwargs)
        if scheme == 'https':
            from twisted.internet import ssl
            if contextFactory is None:
                contextFactory = ssl.ClientContextFactory()
            if self.use_proxy:
                ip = yield self.resolve(self.proxy_host)
                reactor.connectSSL(ip, self.proxy_port,
                                   factory, contextFactory)
            else:
                ip = yield self.resolve(self.host)
                reactor.connectSSL(ip, self.port,
                                   factory, contextFactory)
        else:
            if self.use_proxy:
                ip = yield self.resolve(self.proxy_host)
                reactor.connectTCP(ip, self.proxy_port, factory)
            else:
                ip = yield self.resolve(self.host)
                reactor.connectTCP(ip, self.port, factory)

        result = yield factory.deferred
        defer.returnValue(result)


@defer.inlineCallbacks
def getPage(url, concurrent_key=None, concurrent_limit=None, max_retries=1, description=None, *args, **kwargs):
    if description is None:
        description = url
    kwargs['description'] = description

    # Limit concurrent queries if concurrent_key and concurrent_limit are specified.
    semaphore = None
    if concurrent_key is not None and concurrent_limit is not None:
        semaphore = getDeferredSemaphore(concurrent_key, concurrent_limit)

    factory = ProxyWebClient(url)

    def sleep(secs):
        d = defer.Deferred()
        reactor.callLater(secs, d.callback, None)
        return d

    # Incremental backoff as outlined by AWS.
    # http://aws.amazon.com/articles/1394
    for retry in xrange(max_retries + 1):
        if retry > 0:
            delay = (random.random() * pow(4, retry)) / 10.0
            log.debug(
                '%s retry %s backoff is %s seconds',
                description, retry, delay)

            yield sleep(delay)

        try:
            if semaphore:
                result = yield semaphore.run(factory.get_page, *args, **kwargs)
            else:
                result = yield factory.get_page(*args, **kwargs)

        except DNSLookupError, e:
            if retry < max_retries:
                # These could be transient, so go ahead and retry.
                log.debug("Retrying after DNS Error: %s" % str(e))
                continue
            else:
                raise e

        except (ConnectionClosed, TimeoutError) as ex:
            raise ex

        except Exception, ex:
            code = getattr(ex, 'status', None)

            try:
                # Try to improve the message, based on response XML, if possible.
                # (ignore the xml namespace)
                xml_str = ex.response.replace(' xmlns=', ' xmlnamespace=', 1)
                xml = etree.parse(StringIO(xml_str))
                ex.response = "%s: %s" % (
                    xml.xpath('//Error/Code/text()')[0],
                    xml.xpath('//Error/Message/text()')[0]
                )
            except Exception:
                pass

            if code in ('500', '503'):
                continue
            if code in ('400',):
                # A 400 error is pretty unusual, and probably not transient,
                # but it can't hurt to try it again, just in case.

                try:
                    error_message = "%s %s: %s" % (code, ex.message, ex.response)
                except AttributeError:
                    error_message = str(ex)

                if "AuthorizationHeaderMalformed" in error_message:
                    # no point in retrying one of these..
                    log.debug("HTTP Error (%s): %s" % (url, error_message))
                    raise

                if retry < max_retries:
                    log.debug("HTTP Error (%s): %s - retrying (%d/%d)." % (url, error_message, retry, max_retries))
                    continue
                else:
                    # raise the exception after the max number of retries is exceeded.
                    if max_retries > 1:
                        log.debug("HTTP Error after %d retries: (%s): %s" % (max_retries, url, error_message))
                    else:
                        log.debug("HTTP Error: (%s): %s" % (url, error_message))
                        raise ex

                continue

            log.debug(
                "Unhandled error has been occurred in getPage method "
                "while connecting to %s, %s",
                url, str(ex))

            # any other exception, we can't recover, so give up.
            raise ex

        else:
            defer.returnValue(result)
            break


def getDeferredSemaphore(key, limit):
    if key not in SEMAPHORES:
        SEMAPHORES[key] = defer.DeferredSemaphore(limit)
    semaphore = SEMAPHORES[key]

    if semaphore.limit != limit:
        if limit >= semaphore.tokens:
            semaphore.limit = limit
            log.info("Unable to lower maximum parallel query limit for %s to %d ", key, limit)
        else:
            log.warning("Unable to lower maximum parallel query limit for %s to %d at this time (%d connections currently active)", key, limit, semaphore.tokens)

    return semaphore


def getSESRegions():
    addLocalLibPath()
    import boto.ses
    region_infos = boto.ses.regions()
    regions = []
    for region in region_infos:
        regions.append(region.name)
    return regions


def updateToMany(relationship, root, type_, ids):
    '''
    Update ToMany relationship given search root, type and ids.

    This is a general-purpose function for efficiently building
    non-containing ToMany relationships.
    '''
    root = root.primaryAq()

    new_ids = set(map(prepId, ids))
    current_ids = set(o.id for o in relationship.objectValuesGen())
    changed_ids = new_ids.symmetric_difference(current_ids)

    query = Or(*(Eq('id', x) for x in changed_ids))

    obj_map = {}
    for result in ICatalogTool(root).search(types=[type_], query=query):
        obj_map[result.id] = result.getObject()

    for id_ in new_ids.symmetric_difference(current_ids):
        obj = obj_map.get(id_)
        if not obj:
            continue

        if id_ in new_ids:
            relationship.addRelation(obj)
        else:
            relationship.removeRelation(obj)

        # Index remote object. It might have a custom path reporter.
        notify(IndexingEvent(obj, 'path', False))

        # For componentSearch. Would be nice if we could target
        # idxs=['getAllPaths'], but there's a chance that it won't exist
        # yet.
        obj.index_object()


def updateToOne(relationship, root, type_, id_):
    '''
    Update ToOne relationship given search root, type and ids.

    This is a general-purpose function for efficiently building
    non-containing ToOne relationships.
    '''
    old_obj = relationship()

    # Return with no action if the relationship is already correct.
    if (old_obj and old_obj.id == id_) or (not old_obj and not id_):
        return

    # Remove current object from relationship.
    if old_obj:
        relationship.removeRelation()

        # Index old object. It might have a custom path reporter.
        notify(IndexingEvent(old_obj.primaryAq(), 'path', False))

    # No need to find new object if id_ is empty.
    if not id_:
        return

    # Find and add new object to relationship.
    root = root.primaryAq()
    query = Eq('id', id_)

    for result in ICatalogTool(root).search(types=[type_], query=query):
        new_obj = result.getObject()
        relationship.addRelation(new_obj)

        # Index remote object. It might have a custom path reporter.
        notify(IndexingEvent(new_obj.primaryAq(), 'path', False))

        # For componentSearch. Would be nice if we could target
        # idxs=['getAllPaths'], but there's a chance that it won't exist
        # yet.
        new_obj.index_object()

        return


def prodState(state):
    # if the state is stopped return decommissioned, otherwise production.
    if state == 'stopped':
        return -1
    else:
        return 1000


def twisted_web_client_parse(url, defaultPort=None):
    """
    Split the given URL into the scheme, host, port, and path.

    @type url: C{str}
    @param url: An URL to parse.

    @type defaultPort: C{int} or C{None}
    @param defaultPort: An alternate value to use as the port if the URL does
    not include one.

    @return: A four-tuple of the scheme, host, port, and path of the URL.  All
    of these are C{str} instances except for port, which is an C{int}.
    """
    url = url.strip()
    parsed = http.urlparse(url)
    scheme = parsed[0]
    path = urlunparse(('', '') + parsed[2:])

    if defaultPort is None:
        if scheme == 'https':
            defaultPort = 443
        else:
            defaultPort = 80

    host, port = parsed[1], defaultPort
    if ':' in host:
        host, port = host.split(':')
        try:
            port = int(port)
        except ValueError:
            port = defaultPort

    if path == '':
        path = '/'

    return scheme, host, port, path


def format_time(time):
    '''
    Return formatted time string.
    '''
    return time[:time.rfind('.')].replace('T', ' ')
